"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                      RAISING EXCEPTIONS                                     ||
      ||                                       (run-time errors)                                     ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   The phrase "raising an exception" means generating a run-time error. Normally, an error aborts a run. But
it is possible to catch or trap these exceptions (errors) and keep on running. A program that does this is
called "robust", and catching exceptions is an example of defensive programming. But why would you want your
program to keep on running with a big error? The answer is that a) sometimes you can include the code that
will fix the error or b) you can warn the user during the run. The following example catches bad input. The
code repeatedly tells the user to fix the input.

# EXAMPLE 1: (Insist input is an integer.)
    flag = True
    s = 'Enter an integer:'
    while flag:
       try:
          num = int(input(s))
       except:
          s = 'That was NOT an integer! Try again.\nEnter an integer: '
       else:
          flag = False
    print ('num =', num)
#-------------------------------------------------------------
# EXAMPLE 2: (The exception was caught and the program continues to run.)
    try:
       x = 3/0
    except:
       print ('ERROR: Division by zero. The variable x was set to 5.')
       x = 5
    print('x =', x)
# Output:
#        ERROR: Division by zero.
#        x = 5

   Suppose we requested a radius r and a center (x,y) for a circle. What happens if the user submits a
negative radius? Should we use absolute value to make the radius positive? This is the solution chosen by the
TIB (T.I. BASIC) designer. Or should we ask for a re-submission of the radius value? It depends on the
application. Defensive programming can be misused. But why catch an exception if you are going to abort the
program. Instead use an assertion.

# EXAMPLE 3: (Using a assert instead of an exception.)
    x = 3
    assert 4 < x < 10, 'Error: x is out of bounds.'
# Output:
# Traceback (most recent call last):
# ...
# AssertionError: Error: x is out of bounds.
#-------------------------------------------------------------

 EXAMPLE 4: (Use an assertion instead.)
    x = 3
    if x < 4 or x > 10:
        raise Exception ('Error: x is out of bounds')
#-------------------------------------------------------------

# EXAMPLE 5: (Use an assertion instead.)
       x = 3
       if 4 < x < 10:
          exit('Error: x is out of bounds.')
#-------------------------------------------------------------
# EXAMPLE 6: (Use an assertion instead.)
    x = 3
    if 4 < x < 10:
        print ('Error: x is out of bounds.')
        x = x/0

