"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                       LISTS AND TUPLES                                      ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   What are called ARRAYS in C, C++, and Java are called (mutable) LISTS in Python--e.g., L = [1,2,3,] (note
the square brackets).

  THE LIST METHODS AND OPERATORS:

  L = ['A', 'B', 'C', 'D', 'E', 'F',]

  1. L[n] = 'X'       # MODIFIES the nth element of a list. The first element in a list L is L[0]. The final
                      # element is L[-1] =  L[len(L)-1]

  2. L.append('G')    # APPENDS a SINGLE element only 'G' to END of list. L.append(['G', 'H']) will append the
                      # list ['G', 'H']. L[len(L):] = ['G', 'H'] does the same. But the two instructions
                      # L += ['G', 'H'], and L.extend(['G', 'H']) both append the two INDIVIDUAL elements 'G'
                      # and 'H' to the end of the given list. Unfortunately, L+= ... will create an entirely
                      # new list. So append and extend are better. NOTE WELL: append returns None. Thus,
                      # M = Q.append('G') is the same as M = None. Beware.

  3. L.count('C')     # Returns the number of OCCURANCES of a given element in a given list. If no
                      # occurrences, then a zero is returned.

  4. del(L[n])        # removes the nth element from list. The slice operation L = L[:n]+ L[n+1:] also removes
                      # the nth element from the list. Slicing can also be used to append elements.

                            L = [1, 2, 3, 4, 5]
                            del(L[0]) # in-place deletion
                            print(L)  # output: [2, 3, 4, 5]
                            del(L[:2])
                            print(L)  # output: [4, 5]
                            L = [1, 2, 3, 4, 5]
                            L[2:] = []
                            print(L)  # output: [1, 2]

  5. enumerate(L)     # Places the elements in numbered TUPLES. Here is some code that uses reverse().
                      # Reverse() will not work on tuples, of course.
                        L = ['A', 'B', 'C']     # The given list
                        L = list(enumerate(L))  # [(0, 'A'), (1, 'B'), (2, 'C')]
                        L = list(map(list, L))  # [[0, 'A'], [1, 'B'], [2, 'C']]
                        [x.reverse() for x in L]# [['A', 0], ['B', 1], ['C', 2]]
                        L = list(map(tuple, L)) # [('A', 0), ('B', 1), ('C', 2)]

  6. L.extend('G')    # APPENDS the given element to end of the given list. L.extend(['G', 'H']) will append
                      # the INDIVIDUAL elements (not the single list) to the end of the given
                      # list.

  7. L.index('C')     # Returns the position (index) of the given element in a given list. WARNING: If the
                      # element is not in the list an error is generated.

  8. L.insert(n, 'X') # INSERTS an element in front of the nth element of the given list. This is the same
                      # as L[n:n] = ['X'].

  9. len(L)           # returns the number of elements in list L.

 10. L.pop()          # REMOVES and RETURNS the element at the back of the list ('F', above). An empty list
                      # gives an error.

 11. L.pop(n)         # Removes and returns the element with INDEX n. L.pop(1) would remove and return 'B'.

 12. L.remove('C')    # REMOVES first occurrence of element in list. WARNING: If the element is not in the
                      # list, then an error is generated.

 13. L.reverse()      # REVERSES the given list in place. It returns None. L = L[::-1] also reverses a list.
                      # WARNING: reverse() operates only on lists, not tuples.

 14 reversed()        # Returns a reversed copy.
                                L = [1, 2, 3, 4]
                                M = list(reversed(L))
                                print(M) # output: [4, 3, 2, 1]

 15. L = L[n:]+L[:n]  # ROTATES a list. If n = 2, the result is ['C', 'D', 'E', 'F', 'A', 'B']

 16. L.sort()         # SORTS the list in place. All elements must be of same type. The sort function will
                      # sort lists of lists by comparing their initial elements, then their next elements,
                      # etc. See the section on sorting for more information.

                                L = [1, -2, 3, -4]
                                L.sort(key = abs, reverse = True)
                                print(L) # output: [-4, 3, -2, 1]

                      # A common error is to write L = L.sort(), which sets L to None.

 17. L = A.sorted()   # Returns a sorted list WITHOUT MODIFYING the original list.

 18. 'X' in L         # MEMBERSHIP (boolean). You can also write 'X' not in A.

 19. "+"              # CONCATINATION ['A','B'] + ['C','D'] =['A','B','C','D']
                      # WARNING: You can NOT add matrices componentwise with the '+'. Instead, you need to
                      # overload the '+' operator with __add__ (self, x, y). See the section on classes.

                      # WARNING: Do not ever use "+" with lists--e.g.,  Lst = Lst + [1] is bad, because it
                      # needlessly reconstructs the entire list in in memory (slow) and breaks passing-by-
                      # reference in functions. Instead, use Lst.append(1).

 20. L = [0] * 5      # REPLICATION operator: L = [0, 0, 0, 0, 0], but beware of [[0] * 5] *3 which will claim
                        only 5 places in memory, not 15 places.

 21. min([7,3,1,0,2]) # Finds the MINIMUM element in a list of all same types.

 22. max([7,3,1,0,2]) # Finds the MAXIMUM element in a list of all same types.

 23. L = A[:]         # Both work for COPIES a list, but NOT a list of lists.
     L = list(A)      # For a list of lists. Use deepcopy for that.

 24. deepcopy         # This is the ONLY way to copy a list of lists:

                                from copy import deepcopy
                                A = deepcopy (B)
#-------------------------------------------------------------------------------------------------------------

   NOTE: list-type parameters are passed by REFERENCE, not VALUE. Any change to the received variable is
made to the passed variable. This is usually a good idea, because it saves memory and time. But if you need
the variable passed by value, then make a copy immediately. Here is an example.

                            def change(Lst):
                                Lst[0] = 'Z'
                                Lst = Lst[:]
                                Lst [1] = 'Z'
                            def main():
                                Lst = ['A', 'B', 'C', 'D',]
                                change (Lst)
                                print(Lst) # output: ['Z', 'B', 'C', 'D']

   Beware of using "+" with lists. Your list that was passed by reference may suddenly be a copy at a new
place in memory.

                    A = [1]
                    B = C = A      # A, B, and C are the same place in memory.
                    A = A + [2]    # A is reconstructed at a new memory address.
                    C += [3]       # C and B are both changed at the old memory address.
                    print(A, B, C) # output: [1, 2] [1, 3] [1, 3]


   Avoid defaulting to a mutable expression, because this expression is given only one place in memory, and is
not allowed to change its location. Consequently, when you call the function again, the default value is NOT
used to re-initialize.
                            def doit1(x = None): # Correct
                                if x == None:
                                    x = []
                                x.append(1)      # Recall: x.append(1) appends in place and returns None.
                                return x

                            def doit2(x = []):   # A BIG mistake
                                x.append(1)
                                return x

                            print ( doit1() ) # Output: [1]
                            print ( doit1() ) # Output: [1]
                            print ( doit2() ) # Output: [1]
                            print ( doit2() ) # Output: [1, 1] <-- WHAT?!
#-------------------------------------------------------------------------------------------------------------

                            # Be careful. The following code will NOT change M2.
                                for row in M2:
                                    for cell in row: # Mistake
                                        cell = 0

                            # Here is the code to change M2.
                                for r in range(len(M1)):
                                    for c in range (len(M1[0])):
                                        M2[r][c] = 0
#-------------------------------------------------------------------------------------------------------------
   Here is how to round the elements in a list:

                                Lst = [1.99, 1.11, 2.85, 2.24,]
                                Lst  = [round(x, 1) for x in Lst]
                                print (Lst) # Output: [2.0, 1.1, 2.9, 2.2]
#-------------------------------------------------------------------------------------------------------------
   Here is a handy trick for subscripting a list.

    def fib (n):
        if n < 19: return [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610, 987, 1597, 2584,][n]
        return fib(n-1) + fib(n-2)
#-------------------------------------------------------------------------------------------------------------

   Subsets of lists can be obtained by slicing. NOTE: L[-1] =  L[len(L)-1]. Using the colon is called
"slicing". The general form for slicing is L[i:j:stride]. Indexing out of bounds raises an error. But slicing
out of bounds is acceptable. Here are some examples.

   1. L = ['A', 'B', 'C', 'D', 'E', 'F',]
   2. print(L[1])     # output: B
   3. print(L[1:3])   # output: ['B', 'C'] <-- Note the quote marks.
   4. print(L[:3])    # output: ['A', 'B', 'C'], 0 is the default first argument.
   5. print(L[3:])    # output: ['D', 'E', 'F'], includes final element.
   6. print(L[:])     # ['A', 'B', 'C', 'D', 'E', 'F'], includes every element.
   7. print(L[-1])    # output: F
   8. print(L[-2:-1]) # output: ['E']
   9. print(L[-3:-1]) # output: ['D', 'E']
  10. print(L[-3:])   # output: ['D', 'E', 'F']
  11. print(L[-2:3])  # output: []
  12. print(L[1:4: 2])# output: ['B', 'D'], stride forward by 2
  13. print(L[1:4:-2])# output: []
  14. print(L[::-1])  # output: ['F', 'E', 'D', 'C', 'B', 'A'], stride back by 1
  15. print(L[::-2])  # output: ['F', 'D', 'B'], stride back by 2
  16. print(L[  :-2]) # ['A', 'B', 'C', 'D']
  17. print(L[ 2:-2]) # ['C', 'D']
  18. print(L[-2:-4]) # []
  19. print(L[-100:100])# ['A', 'B', 'C', 'D', 'E', 'F'], no error
  20. for x in range(10)[::2]:
        print(x, end = ", ") # output: 0, 2, 4, 6, 8,
#-------------------------------------------------------------------------------------------------------------

   In Python, lists are arrays (continuous blocks of memory, with offsets), but they act as if they were
linked-lists, hence the name.  Consequently, adding and removing elements (except at the end of a list) is
inefficient. However, Python does have a Queue type, which is a true linked-list type. See the section on
QUEUES.

Remember that C.S. mainly uses three kinds of queues:

           1. FIFO (first in, first out)
           2. LIFO = stack (last in, first out), and
           3. priority, based on some ordering function.

        from queue import * # Capitalized as Queue in Python versions 2.x.
        def main():

        #---FIFO Queue: push(bottom) and pop(top)
            q = Queue() # <-- This is a FIFO queue.
            Lst = ['A','B','C','D']
            for elt in Lst:
                q.put(elt)       # insert at back
            print ( "Size = ", q.qsize() )      # output: Size =  4
            print(Lst)                          # output: ['A', 'B', 'C', 'D']
            while not q.empty():
               print (q.get(), end = "")        # output: ABCD
            print()
            print ( "Size = ", q.qsize() )      # output: Size =  0

It is common to place a comma after the final element in a list (or tuple)--e.g.,

                       L = ['A', 'B', 'C', 'D, 'E', 'F',]

Note well: the elements of a list do not have to be of the same data type--e.g., the following list is legal.

          list1 = ['string', 3.2, 5, ['sub', 'list', 'of', 'strings']]

   A common error is to assign a value to a position in an empty list. Consequently, we sometimes build a list
of zeroes or Nones, so there is a place in memory to which we can assign a values. The single line of code
below creates a list of 100 Nones. As a place holder, None is preferred to zero. Why? Because zero is a value,
and it is of a particular type. The value None is more generic. It is, however, a data type, but a data type
with only one value.

                         L = [None for x in range(100)]

  Note that you can append items to a list by using

     a) L.append('X') at end of list for a SINGLE item (best!).

     b) L.extend(['X', 'Y']) at end of list for several items to be inserted (also good).

     c) L.insert(n, 'X') at position n for a single item. (BAD!) NOTE WELL: inserting in the middle of a long
        list or at the front is inefficient.

     d) L += ['X', 'Y'] at end of list for several items. (This is the WORST, method, because it creates a new
        list in memory, instead of just appending to the old list. Note Well: The "+=" can lead to an error in
        recursion.)

  You can remove items from a list using the following instructions. However, removing from anywhere but the
end of a list is slow, because shifting is done behind the scenes.

     a) del(L[n]) best, if you know the index.
     b) pop(n) removes and returns item at index n, if list is not empty.
     c) L.remove('C'), removes the first occurrence of 'C'. If the element is not there, an error occurs.

   A handy attribute of lists is that a non-empty list returns True, and an empty list returns False. For a
discussion of lists as queues and the queue class see the section on queues. [Recall the queues: LIFO (aka a
stack) FIFO (first in, first out), and a priority queue (a sorted queue).


   WARNING: You can NOT copy a simple list like this: L = A, because both identifiers refer to the same place
in memory. Instead use one of these methods:

                          L = A[:] or L = list(A).

   But even these methods will FAIL with a list that contains lists--e.g., a doubly subscripted list (AKA a
matrix, a table). The only way to create a copy that is unconnected to the original list is like this:

                        from copy import deepcopy
                        L = deepcopy(A)

   The term "deep" means that lists of lists of lists, etc. are copied by value, not reference. L = list(A)
gives you only a "top-level" copy. Most of the time you want to work with the original list (make no copies,
to save on memory) or a top-level copy will give you what you want. But matrices require deepcopy.

   We must be careful of lists of lists--i.e., doubly-subscripted arrays, AKA matrices. Since L = [0]*4 gives
us [0,0,0,0], you would think that L = [[0]*5] * 3] would give us [[0,0,0,0],[0,0,0,0],[0,0,0,0]].  And if you
run the code, and print it, that is what you will see. But there is a horrible trap built in. You are looking
at the same row in memory just printed four times. To change L[0][0] is to change L[1][0] and change L[2][0].
Yikes!

    L = [0] * 4
    L[0] = 1          # OK
    print (L)         # output: [1, 0, 0, 0]

    L = [[0] * 4] * 3
    L[0][0] = 1       # BAD NEWS!
    print (L)         # output: [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0]]


    L = [[0 for col in range(3)] for row in range(4)] # This is the best way.
    L[0][0] = 2       # OK
    print(L)          # output: [[2, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]

    Both lists and tuples can be zipped into a single set of tuples.
    x = [1, 2, 3] # or x = (1, 2, 3)
    y = [4, 5, 6] # or y = (4, 5, 6)
    zipped = list(zip(x, y)) # or tuple(zip(x, y))
    print ("zipped = ", zipped) # zipped =  [(1, 4), (2, 5), (3, 6)]

    Now, meet the star operator(*), which looks like the multiplication sign, but is placed in front of a
parameter, list or tuple. The star operator can both pack and unpack lists and tuples. Below the star operator
is used to unpack a tuple and unpack a list:

    L = (1,2,3)     # packed as a tuple
    print (L)       # output: (1 2 3) a tuple (still packed)
    print (*L)      # output: 1 2 3   separate items (tuple unpacked)
    print (list(L)) # output: [1 2 3] a list (packed)
    print (*list(L))# output: 1 2 3   separate items (list unpacked)

   The most important use of the star operator is to pack an unspecified number of parameters received in a
function into a tuple (polymorphism).

   def doit (*data)     # data = (1,2,3, ...) packed into a tuple

   So the star is 1) a tuple packer (in a header), 2) a list unpacker, and 3) a tuple unpacker.

   Below are two functions that turn a positive integer into a list of digits and vice versa.

            def changePositiveIntegerToList(num):
               L = []
               for n in range(len(str(num))):
                  L.append(num%10)
                  num = num//10
               L.reverse() # reverse in place, nothing is returned.
               return L


            def changeIntegerListToInteger(L):
               num = 0
               L.reverse()
               for n in range(len(L)):
                   num += L[n]*10**n
               return num

            num = 19354
            L = changePositiveIntegerToList(num)
            print (L)  # output: [1, 9, 3, 5, 4]
            num = changeIntegerListToInteger(L)
            print(num) # output: 19354
#-------------------------------------------------------------------------------------------------------------

                                                        TUPLES

   Tuples are almost the same as lists: they are ordered and indexed, slicing and negative indices work with
tuples. The differences are 1) tuples are immutable (you can't append, remove, sort, or change their
elements), they have only a few built-in functions--e.g., append() and reverse() will NOT work with tuples.
2) tuples use parentheses, not square brackets. But sometimes the parentheses are optional. See below.

                        a, b, c = (10,20,30)
                        print(b)          # output: 20
                        a, b, c = range(3)
                        print((a,b,c,))   # output: (0, 1, 2)

   In the first line above (line a, b, c = (10,20,30)) we say the tuple on the right has been UNPACKED. The
parentheses on the left-hand tuple (a, b, c) are optional and here have been omitted. Below is an example of
unpacking.

    circle = (3, 4, 5)
    pos =  dist ( circle[:2]) # passes to dist() the ONE parameter (3, 4).
    pos =  dist (*circle[:2]) # passes to dist() the TWO parameters 3, 4.

   The advantages of tuples are

    a) tuples are FASTER than lists
    b) tuples write-protect data
    c) tuples are commonly used for dictionary keys.
    d) tuples, as well as lists, allow functions to return multiple values.
    e) tuples allow multiple assignments in one line. See code below.

   Modifying a tuple can be done in two ways.

                #   Method 1. type casting (inefficient)
                    T = (1,2,3,4,5)
                    T = list(T)
                    T[3] = 7
                    T = tuple(T)
                    print (T) # output: (1, 2, 3, 7, 5)

                #   Method 2. slicing (best)
                    T = (1,2,3,4,5)
                    T = T[:3] + (7,) + T[4:]
                    print (T) # output: (1, 2, 3, 7, 5)

   WARNING: One-element tuples require a comma. I suggest you get into the habit of ending almost every tuple
(and list) with a comma.

   Both lists and tuples can be zipped:

    L1 = [2,4,6,8]
    L2 = [7,5,3,1]
    for (x,y) in zip(L1, L2):
       print (x, '+', y, '=', x+y, end = ', ')
    #  output: 2 + 7 = 9, 4 + 5 = 9, 6 + 3 = 9, 8 + 1 = 9,


   WARNING: When the 3rd element is removed, the list length falls to 3, and the for loop ends without looking
at the fourth element. The saving trick is to use a copy with "[:]".

# A BIG MISTAKE
#---Example 6. Modifying the range. (This modification CAN be done, and is usually a mistake.)
    Lst = [1, 2, 3, 4, 5, 6,]
    for x in Lst: # This should be "for x in Lst[:]:"--i.e., make a copy of Lst.
          if  x > 2:
            Lst.remove(x)
    print(Lst) # output: [1, 2, 4, 6], an unexpected result.

#---Example 7. Modifying the range.
    Lst = [1, 2, 3, 4, 5, 6,]
    for x in Lst[:]:
          if  x > 2:
            Lst.remove(x)
    print(Lst) # output: [1, 2], the expected result
