"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                        RANDOM NUMBERS                                       ||
      ||                    References: David M. Beazley, Python Essential Reference,                ||
      ||                                4th Ed. (Addison Wesley, 2009), pages 254-256.               ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

   WARNING: Do not name your program "random". If you do, the built-in random function will raise exceptions
whenever it is called. The function random() returns a random float between 0 and 1, including 0, but not
including 1.
                                   from random import random
                                   r = random() # 0 <= r < 1

   The period of this random number generator (before it starts repeating) is 2**19937-1.
#-------------------------------------------------------------------------------------------------------------

  The following are a few of the Python built-in functions.

   r = choice (['A', 'B', 'C'])
   r = choice ('ABC') # Same as above.

1. from random import random
   r = random()                 # 0 <= r < 1
   r = 5;
   if random() > 0.75: r = 17 # Randomly choose 5 with a probability of 0.75, otherwise choose 17.

2. from random import choice
   r = choice ([1, 13, 7, 12])  # All have equal probability of being chosen.
   r = choice (['A', 'B', 'C'])
   r = choice ('ABC') # Same as above.

3. from random import randint
   r = randint(a, b)            # a <= r <= b

4. from random import randrange
   r = randrange(a, b, s)       # a <= r < b --> a, a+s, a+2s, a+3s, etc.
   r = randrange(3, 7, 2) # Randomly choose an ODD integer from 3 to 7 (NOT including 7).

5. from random import shuffle
   seq = ['A', 'B', 'C', 'D', 'E', 'F']
   shuffle(seq)                 # shuffle in place

6. from random import sample
   sample([1, 2, 3, 4, 5],  3)  # [4, 2, 5]

7. Print all 2-letter permutations of letters chosen from 'ABC'.
   Note: ('A', 'B') is NOT the same as ('B', 'A').
   from itertools import permutations
   L = list(permutations('ABC', 2)) # [('A', 'B'), ... ('C', 'B')], etc.
   print ('L =', L)

8. Randomly generate the two numbers 1 and -1. Some methods of generating random numbers are better than
   others.
         +---------------------------------------------------------+
         |  1.  r = 1                                              |
         |      if random() > 0.5: r = -1    # 0.21 milli-seconds  |
         |                                                         |
         |  2.  r = int(random() > 0.5)*2-1  # 0.62 milli-seconds  |
         |                                                         |
         |  3.  r = int(random()*2)*2 -1     # 0.81 milli-seconds  |
         |                                                         |
         |  4.  r = (-1)** int(random()*2)   # 0.98 milli-seconds  |
         |                                                         |
         |  5.  from random import choice                          |
         |      r = choice ([1, -1])         # 1.57 milli-seconds  |
         |                                                         |
         |  6.  from random import randrange                       |
         |      r = randrange(-1, 2, 2)      # 3.26 milli-seconds  |
         +---------------------------------------------------------+
#-------------------------------------------------------------------------------------------------------------

    Weighted random choices are occasionally needed. For example, in this list Lst = [1, 0, 3, 0, 9, 2], the
probability of choosing the 9 should be 9/15. The probability of choosing either 0 should be 0/15 = 0. This is
a problem that commonly arises in artificial intelligence. So how do we solve it? Here are two solutions.

                  +--------------------------------------------------+
                  |     def randCell1(Lst):                          |
                  |          from random import randint              |
                  |          r = randint(1, sum(Lst))                |
                  |          t = 0                                   |
                  |          for cell in Lst:                        |
                  |              t += cell                           |
                  |              if t >= r: return cell              |
                  |          exit('error in randCell1')              |
                  |--------------------------------------------------+
                  |                                                  |
                  |     def randCell2(Lst):                          |
                  |         from itertools import accumulate         |
                  |         from random import randint               |
                  |         r = randint(1, sum(Lst))                 |
                  |         for n, x in enumerate(accumulate(Lst)):  |
                  |             if r <= x: return (Lst[n])           |
                  |         exit('error in randCell2')               |
                  +--------------------------------------------------+
#-------------------------------------------------------------------------------------------------------------

   If we need to generate all possible permutations of a list. Here is the built-in function to do the job:

            from itertools import *
            print ("All permutations:", list(permutations([1,2,3,4], 4)))

   The pretty printed output is below.

    +-------------------------------------------------------------------------+
    |  0 [1, 2, 3, 4]    6 [2, 1, 3, 4]    12 [3, 1, 2, 4]   18 [4, 1, 2, 3]  |
    |  1 [1, 2, 4, 3]    7 [2, 1, 4, 3]    13 [3, 1, 4, 2]   19 [4, 1, 3, 2]  |
    |  2 [1, 3, 2, 4]    8 [2, 3, 1, 4]    14 [3, 2, 1, 4]   20 [4, 2, 1, 3]  |
    |  3 [1, 3, 4, 2]    9 [2, 3, 4, 1]    15 [3, 2, 4, 1]   21 [4, 2, 3, 1]  |
    |  4 [1, 4, 2, 3]   10 [2, 4, 1, 3]    16 [3, 4, 1, 2]   22 [4, 3, 1, 2]  |
    |  5 [1, 4, 3, 2]   11 [2, 4, 3, 1]    17 [3, 4, 2, 1]   23 [4, 3, 2, 1]  |
    +-------------------------------------------------------------------------+
#-------------------------------------------------------------------------------------------------------------

   Occasionally we need to generate the rth permutation from some ordering of a sequence (n-choose-r = nPr).
This way we can generate one permutation at a time and never encounter the same permutation twice. Here is
such a function .

        +---------------------------------------------------------+
        |   def permute(Lst, r):                                  |
        |    #--Work with copy.                                   |
        |       Lst = Lst[:]                                      |
        |    #--Warn of parameter mistake (robust).               |
        |       L = len(Lst)                                      |
        |       if L<1 or r<0 or r>=factorial(L) \                |
        |              or type(Lst)!=list or type(r)!=int: \      |
        |              print ('ERROR: bad data in def permute.')  |
        |              return Lst                                 |
        |    #--base case                                         |
        |       if L == 1: return Lst                             |
        |    #--recursive case                                    |
        |       d     = factorial(L-1)                            |
        |       digit = Lst[r//d]                                 |
        |       Lst.remove(digit)                                 |
        |       return [digit] + permute(Lst, r%d)                |
        +---------------------------------------------------------+
"""