"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                           RECURSION                                         ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   Recursive functions are always described as functions that call themselves. Actually they push values on a
memory stack. Generally, global variables (not constants) are to be avoided. Why? Keeping the scope of a
variable as short as possible reduces the probably of errors. However, global variables in recursion often
lead to simplification of the code. Since recursion is often so difficult, the simplification seems worth
breaking with the guideline.

   It is possible to do recursion by having function A call function B, which in turn calls function A. This
trick , called cyclic recursion, is usually to be avoided. Curiously, Python has no built-in function to
flatten a list. It is easy to write a recursive flatten function.

"""
from sys import setrecursionlimit
setrecursionlimit(100) # optional (default = 1000)

def fib (n): # simple
  if n < 1: return None
  if n == 1 or n == 2: return 1
  return fib(n-1) + fib(n-2)
#-------------------------------------------------------------------------------

def fib (n): # robust and fast
  if type(n) != int or n < 1:
    exit('Error in fib parameter') # <-- Checking data makes a robust function.
  if n < 27: return FIB[n]         # <-- base case (fast and less memory needed)
  return fib(n-1) + fib(n-2)       # <-- recursive case

#--The global below contains the first 26 Fibonacci numbers.
FIB = [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610, 987, ] + \
      [1597, 2584, 4181, 6765, 10946, 17711, 28657, 46368, 75025, 121393,]
#-------------------------------------------------------------------------------

def sum2 (L):
  return 0 if not L else L[0] + sum2(L[1:]) # ternary if-else
#-------------------------------------------------------------------------------

def sum3 (L):
  top, *bottom = L
  return top if not bottom else top + sum2(bottom)
#-------------------------------------------------------------------------------

def sum4 (top, *bottom): # BAD example
   print (top)   # top = [1, 2, 3, 4]
   print (bottom)# bottom = ()
   return top if not bottom else top + sum2(bottom)
#-------------------------------------------------------------------------------

def A (L):
    return 0 if not L else B(L)

def B (L):
    return L[0] + A(L[1:])
#-------------------------------------------------------------------------------

def sum5 (L):
  if not L: return 0
  if type(L[0]) == int:  return      L[0]  + sum5(L[1:])
  if type(L[0]) == list: return sum5(L[0]) + sum5(L[1:])
  print('ERROR 1:', L[0], 'is neither an integer, nor a list.')
#-------------------------------------------------------------------------------

def flatten (L):
  Lst = []
  for elt in L:
     if type(elt) in (list, tuple):
           Lst.extend(flatten(elt))
     else: Lst.append(elt)
  return Lst
#===============================================================================
def main():

#--Example 1. The classic fibonacci sequence.
   print ('1.', fib(5)) # output: 5

#--Example 2-4. Two ways to RECURSIVELY iterate over a sequence (here,
#  summing the elements).
   L = [1,2,3,4,]
   print ('2.', sum2(L)) # output: 10
   print ('3.', sum3(L)) # output: 10
   print ('4.', sum4(L)) # output: [1, 2, 3, 4] <-- failed to sum!

#--Example 5. Pairwise recursive functions.
   print ('5.',    A(L)) # output: 10

#--Example 6. Summing the numbers of lists inside of lists inside of lists, etc.
   L = [1,[2,3],4, [5, [6, 7]], [8, 9, [1,2, [3,4] ] ] ]
   print ('6.', sum5(L)) # output: 55

#--Example 7. Flattening a list of lists, etc.
   L = [1,[2,3],4, [5, [6, 7]], [8, 9, [1,2, [3,4] ] ] ]
   print ('7.', flatten (L))
if __name__ == '__main__': main()
#=============================== END OF PROGRAM ================================